import asyncio
import logging
from typing import List, Dict, Any, Literal, AsyncIterator, Iterator, Optional, Union
from pydantic import PrivateAttr
import cohere

from swarmauri_core.typing import SubclassUnion
from swarmauri.messages.base.MessageBase import MessageBase
from swarmauri.messages.concrete.AgentMessage import AgentMessage
from swarmauri.messages.concrete.HumanMessage import HumanMessage, contentItem
from swarmauri.llms.base.LLMBase import LLMBase
from swarmauri.schema_converters.concrete.CohereSchemaConverter import (
    CohereSchemaConverter,
)


class CohereToolModel(LLMBase):
    """
    A model for interacting with Cohere's API for tool-augmented conversations.
    Provider resources: https://docs.cohere.com/docs/models#command
    """

    api_key: str
    _client: Optional[cohere.Client] = PrivateAttr(default=None)
    allowed_models: List[str] = [
        "command-r",
        "command-r-plus",
        "command-r-plus-08-2024",
    ]
    name: str = "command-r"
    type: Literal["CohereToolModel"] = "CohereToolModel"
    resource: str = "LLM"

    def __init__(self, **data):
        super().__init__(**data)
        self._client = cohere.Client(api_key=self.api_key)

    def model_dump(self, **kwargs):
        dump = super().model_dump(**kwargs)
        return {k: v for k, v in dump.items() if k != "_client"}

    def _schema_convert_tools(self, tools) -> List[Dict[str, Any]]:
        if not tools:
            return []
        return [CohereSchemaConverter().convert(tools[tool]) for tool in tools]

    def _extract_text_content(self, content: Union[str, List[contentItem]]) -> str:
        if isinstance(content, str):
            return content
        elif isinstance(content, list):
            text_contents = [
                item["text"]
                for item in content
                if isinstance(item, dict)
                and item.get("type") == "text"
                and "text" in item
            ]
            return " ".join(text_contents)
        return ""

    def _format_messages(
        self, messages: List[SubclassUnion[MessageBase]]
    ) -> List[Dict[str, str]]:
        formatted_messages = []
        role_mapping = {
            "human": "User",
            "agent": "Chatbot",
            "system": "System",
            "tool": "Tool",
        }

        for message in messages:
            message_dict = {}

            # Extract content
            if hasattr(message, "content"):
                content = message.content
                message_dict["message"] = self._extract_text_content(content)

            # Map role to Cohere expected roles
            if hasattr(message, "role"):
                original_role = message.role.lower()
                message_dict["role"] = role_mapping.get(
                    original_role, "User"
                )  # Default to User if unknown role

            # Add other properties if they exist
            for prop in ["name", "tool_call_id", "tool_calls"]:
                if hasattr(message, prop):
                    value = getattr(message, prop)
                    if value is not None:
                        message_dict[prop] = value

            formatted_messages.append(message_dict)

        return formatted_messages

    def _ensure_conversation_has_message(self, conversation):
        if not conversation.history:
            conversation.add_message(
                HumanMessage(content=[{"type": "text", "text": "Hello"}])
            )
        return conversation

    def _process_tool_calls(self, response, toolkit):
        tool_results = []
        if hasattr(response, "tool_calls") and response.tool_calls:
            for tool_call in response.tool_calls:
                logging.info(f"Processing tool call: {tool_call}")
                func_name = tool_call.name
                func_call = toolkit.get_tool_by_name(func_name)
                func_args = tool_call.parameters
                func_results = func_call(**func_args)
                tool_results.append(
                    {"call": tool_call, "outputs": [{"result": func_results}]}
                )
            logging.info(f"Tool results: {tool_results}")
        return tool_results

    def predict(self, conversation, toolkit=None, temperature=0.3, max_tokens=1024):
        conversation = self._ensure_conversation_has_message(conversation)
        formatted_messages = self._format_messages(conversation.history)
        tools = self._schema_convert_tools(toolkit.tools) if toolkit else None

        tool_response = self._client.chat(
            model=self.name,
            message=formatted_messages[-1]["message"],
            chat_history=(
                formatted_messages[:-1] if len(formatted_messages) > 1 else None
            ),
            force_single_step=True,
            tools=tools,
        )

        tool_results = self._process_tool_calls(tool_response, toolkit)

        agent_response = self._client.chat(
            model=self.name,
            message=formatted_messages[-1]["message"],
            chat_history=(
                formatted_messages[:-1] if len(formatted_messages) > 1 else None
            ),
            tools=tools,
            force_single_step=True,
            tool_results=tool_results,
            temperature=temperature,
        )

        conversation.add_message(AgentMessage(content=agent_response.text))
        return conversation

    def stream(
        self, conversation, toolkit=None, temperature=0.3, max_tokens=1024
    ) -> Iterator[str]:
        conversation = self._ensure_conversation_has_message(conversation)
        formatted_messages = self._format_messages(conversation.history)
        tools = self._schema_convert_tools(toolkit.tools) if toolkit else None

        tool_response = self._client.chat(
            model=self.name,
            message=formatted_messages[-1]["message"],
            chat_history=(
                formatted_messages[:-1] if len(formatted_messages) > 1 else None
            ),
            force_single_step=True,
            tools=tools,
        )

        tool_results = self._process_tool_calls(tool_response, toolkit)

        stream = self._client.chat_stream(
            model=self.name,
            message=formatted_messages[-1]["message"],
            chat_history=(
                formatted_messages[:-1] if len(formatted_messages) > 1 else None
            ),
            tools=tools,
            force_single_step=True,
            tool_results=tool_results,
            temperature=temperature,
        )

        collected_content = []
        for chunk in stream:
            if hasattr(chunk, "text"):
                collected_content.append(chunk.text)
                yield chunk.text

        full_content = "".join(collected_content)
        conversation.add_message(AgentMessage(content=full_content))

    def batch(
        self, conversations: List, toolkit=None, temperature=0.3, max_tokens=1024
    ) -> List:
        results = []
        for conv in conversations:
            result = self.predict(
                conversation=conv,
                toolkit=toolkit,
                temperature=temperature,
                max_tokens=max_tokens,
            )
            results.append(result)
        return results

    async def apredict(
        self, conversation, toolkit=None, temperature=0.3, max_tokens=1024
    ):
        conversation = self._ensure_conversation_has_message(conversation)
        formatted_messages = self._format_messages(conversation.history)
        tools = self._schema_convert_tools(toolkit.tools) if toolkit else None

        tool_response = await asyncio.to_thread(
            self._client.chat,
            model=self.name,
            message=formatted_messages[-1]["message"],
            chat_history=(
                formatted_messages[:-1] if len(formatted_messages) > 1 else None
            ),
            force_single_step=True,
            tools=tools,
        )

        tool_results = self._process_tool_calls(tool_response, toolkit)

        agent_response = await asyncio.to_thread(
            self._client.chat,
            model=self.name,
            message=formatted_messages[-1]["message"],
            chat_history=(
                formatted_messages[:-1] if len(formatted_messages) > 1 else None
            ),
            tools=tools,
            force_single_step=True,
            tool_results=tool_results,
            temperature=temperature,
        )

        conversation.add_message(AgentMessage(content=agent_response.text))
        return conversation

    async def astream(
        self, conversation, toolkit=None, temperature=0.3, max_tokens=1024
    ) -> AsyncIterator[str]:
        conversation = self._ensure_conversation_has_message(conversation)
        formatted_messages = self._format_messages(conversation.history)
        tools = self._schema_convert_tools(toolkit.tools) if toolkit else None

        tool_response = await asyncio.to_thread(
            self._client.chat,
            model=self.name,
            message=formatted_messages[-1]["message"],
            chat_history=(
                formatted_messages[:-1] if len(formatted_messages) > 1 else None
            ),
            force_single_step=True,
            tools=tools,
        )

        tool_results = self._process_tool_calls(tool_response, toolkit)

        stream = await asyncio.to_thread(
            self._client.chat_stream,
            model=self.name,
            message=formatted_messages[-1]["message"],
            chat_history=(
                formatted_messages[:-1] if len(formatted_messages) > 1 else None
            ),
            tools=tools,
            force_single_step=True,
            tool_results=tool_results,
            temperature=temperature,
        )

        collected_content = []
        for chunk in stream:
            if hasattr(chunk, "text"):
                collected_content.append(chunk.text)
                yield chunk.text
            await asyncio.sleep(0)

        full_content = "".join(collected_content)
        conversation.add_message(AgentMessage(content=full_content))

    async def abatch(
        self,
        conversations: List,
        toolkit=None,
        temperature=0.3,
        max_tokens=1024,
        max_concurrent=5,
    ) -> List:
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_conversation(conv):
            async with semaphore:
                return await self.apredict(
                    conv,
                    toolkit=toolkit,
                    temperature=temperature,
                    max_tokens=max_tokens,
                )

        tasks = [process_conversation(conv) for conv in conversations]
        return await asyncio.gather(*tasks)



from pydantic import Field
import asyncio
from typing import List, Literal, AsyncIterator, Iterator
import ai21
from ai21 import AsyncAI21Client
from ai21.models.chat import ChatMessage
from swarmauri_core.typing import SubclassUnion

from swarmauri.messages.base.MessageBase import MessageBase
from swarmauri.messages.concrete.AgentMessage import AgentMessage
from swarmauri.llms.base.LLMBase import LLMBase
from swarmauri.messages.concrete.AgentMessage import UsageData
from swarmauri.utils.duration_manager import DurationManager


class AI21StudioModel(LLMBase):
    """
    Provider resources: https://docs.ai21.com/reference/jamba-15-api-ref
    """

    api_key: str
    allowed_models: List[str] = [
        "jamba-1.5-large",
        "jamba-1.5-mini",
    ]
    name: str = "jamba-1.5-mini"
    type: Literal["AI21StudioModel"] = "AI21StudioModel"
    client: ai21.AI21Client = Field(default=None, exclude=True)
    async_client: AsyncAI21Client = Field(default=None, exclude=True)

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)
        self.client = ai21.AI21Client(api_key=self.api_key)
        self.async_client = AsyncAI21Client(api_key=self.api_key)

    def _format_messages(
        self, messages: List[SubclassUnion[MessageBase]]
    ) -> List[ChatMessage]:
        return [
            ChatMessage(content=message.content, role=message.role)
            for message in messages
        ]

    def _prepare_usage_data(
        self,
        usage_data,
        prompt_time: float = 0,
        completion_time: float = 0,
    ):
        """
        Prepares and extracts usage data and response timing.
        """
        total_time = prompt_time + completion_time

        usage = UsageData(
            prompt_tokens=usage_data.prompt_tokens,
            completion_tokens=usage_data.completion_tokens,
            total_tokens=usage_data.total_tokens,
            prompt_time=prompt_time,
            completion_time=completion_time,
            total_time=total_time,
        )

        return usage

    def predict(
        self,
        conversation,
        temperature=0.7,
        max_tokens=256,
        top_p=1.0,
        stop="\n",
        n=1,
    ):
        formatted_messages = self._format_messages(conversation.history)

        with DurationManager() as prompt_timer:
            response = self.client.chat.completions.create(
                model=self.name,
                messages=formatted_messages,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                stop=stop,
                n=n,
            )

        message_content = response.choices[0].message.content

        usage_data = response.usage

        usage = self._prepare_usage_data(usage_data, prompt_timer.duration)

        conversation.add_message(AgentMessage(content=message_content, usage=usage))

        return conversation

    async def apredict(
        self,
        conversation,
        temperature=0.7,
        max_tokens=256,
        top_p=1.0,
        stop="\n",
        n=1,
    ):
        formatted_messages = self._format_messages(conversation.history)

        with DurationManager() as prompt_timer:
            response = await self.async_client.chat.completions.create(
                model=self.name,
                messages=formatted_messages,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                stop=stop,
                n=n,
            )

        message_content = response.choices[0].message.content

        usage_data = response.usage

        usage = self._prepare_usage_data(
            usage_data,
            prompt_timer.duration,
        )

        conversation.add_message(AgentMessage(content=message_content, usage=usage))

        return conversation

    def stream(
        self,
        conversation,
        temperature=0.7,
        max_tokens=256,
        top_p=1.0,
        stop="\n",
    ) -> Iterator[str]:
        formatted_messages = self._format_messages(conversation.history)

        with DurationManager() as prompt_timer:
            stream = self.client.chat.completions.create(
                model=self.name,
                messages=formatted_messages,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                stop=stop,
                stream=True,
            )

        collected_content = []
        usage_data = {}

        with DurationManager() as completion_timer:
            for chunk in stream:
                if chunk.choices[0].delta.content is not None:
                    content = chunk.choices[0].delta.content
                    collected_content.append(content)
                    yield content

                if hasattr(chunk, "usage") and chunk.usage is not None:
                    usage_data = chunk.usage

        full_content = "".join(collected_content)

        usage = self._prepare_usage_data(
            usage_data, prompt_timer.duration, completion_timer.duration
        )

        conversation.add_message(AgentMessage(content=full_content, usage=usage))

    async def astream(
        self,
        conversation,
        temperature=0.7,
        max_tokens=256,
        top_p=1.0,
        stop="\n",
    ) -> AsyncIterator[str]:
        formatted_messages = self._format_messages(conversation.history)

        with DurationManager() as prompt_timer:
            stream = await self.async_client.chat.completions.create(
                model=self.name,
                messages=formatted_messages,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                stop=stop,
                stream=True,
            )

        collected_content = []
        usage_data = {}

        with DurationManager() as completion_timer:
            async for chunk in stream:
                if chunk.choices[0].delta.content is not None:
                    content = chunk.choices[0].delta.content
                    collected_content.append(content)
                    yield content

                if hasattr(chunk, "usage") and chunk.usage is not None:
                    usage_data = chunk.usage

        full_content = "".join(collected_content)

        usage = self._prepare_usage_data(
            usage_data, prompt_timer.duration, completion_timer.duration
        )

        conversation.add_message(AgentMessage(content=full_content, usage=usage))

    def batch(
        self,
        conversations: List,
        temperature=0.7,
        max_tokens=256,
        top_p=1.0,
        stop="\n",
        n=1,
    ) -> List:
        return [
            self.predict(
                conv,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                stop=stop,
                n=n,
            )
            for conv in conversations
        ]

    async def abatch(
        self,
        conversations: List,
        temperature=0.7,
        max_tokens=256,
        top_p=1.0,
        stop="\n",
        n=1,
        max_concurrent=5,
    ) -> List:
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_conversation(conv):
            async with semaphore:
                return await self.apredict(
                    conv,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    top_p=top_p,
                    stop=stop,
                    n=n,
                )

        tasks = [process_conversation(conv) for conv in conversations]
        return await asyncio.gather(*tasks)



from swarmauri.chains.concrete.CallableChain import CallableChain
from swarmauri.chains.concrete.ChainStep import ChainStep
from swarmauri.chains.concrete.PromptContextChain import PromptContextChain
from swarmauri.chains.concrete.ContextChain import ContextChain

__all__ = [
    "CallableChain",
    "ChainStep",
    "PromptContextChain",
    "ContextChain",
]



import os
import fal_client
import asyncio
from typing import List, Literal, Optional, Union, Dict
from pydantic import Field, ConfigDict
from swarmauri.llms.base.LLMBase import LLMBase


class FalAIVisionModel(LLMBase):
    """
    A model for processing images and answering questions using vision models provided by FalAI.
    Get your API KEY here: https://fal.ai/dashboard/keys
    """

    allowed_models: List[str] = [
        "fal-ai/llava-next",
        "fal-ai/llavav15-13b",
        "fal-ai/any-llm/vision",
    ]
    api_key: str = Field(default_factory=lambda: os.environ.get("FAL_KEY"))
    model_name: str = Field(default="fal-ai/llava-next")
    type: Literal["FalAIVisionModel"] = "FalAIVisionModel"

    model_config = ConfigDict(protected_namespaces=())

    def __init__(self, **data):
        super().__init__(**data)
        if self.api_key:
            os.environ["FAL_KEY"] = self.api_key
        if self.model_name not in self.allowed_models:
            raise ValueError(
                f"Invalid model name. Allowed models are: {', '.join(self.allowed_models)}"
            )

    def _send_request(self, image_url: str, prompt: str, **kwargs) -> Dict:
        """Send a request to the vision model API for image processing and question answering."""
        arguments = {"image_url": image_url, "prompt": prompt, **kwargs}
        result = fal_client.subscribe(
            self.model_name,
            arguments=arguments,
            with_logs=True,
        )
        return result

    def process_image(self, image_url: str, prompt: str, **kwargs) -> str:
        """Process an image and answer a question based on the prompt."""
        response_data = self._send_request(image_url, prompt, **kwargs)
        return response_data["output"]

    async def aprocess_image(self, image_url: str, prompt: str, **kwargs) -> str:
        """Asynchronously process an image and answer a question based on the prompt."""
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(
            None, self.process_image, image_url, prompt, **kwargs
        )

    def batch(self, image_urls: List[str], prompts: List[str], **kwargs) -> List[str]:
        """
        Process a batch of images and answer questions for each.
        Returns a list of answers.
        """
        answers = []
        for image_url, prompt in zip(image_urls, prompts):
            answer = self.process_image(image_url=image_url, prompt=prompt, **kwargs)
            answers.append(answer)
        return answers

    async def abatch(
        self,
        image_urls: List[str],
        prompts: List[str],
        max_concurrent: int = 5,
        **kwargs,
    ) -> List[str]:
        """
        Asynchronously process a batch of images and answer questions for each.
        Returns a list of answers.
        """
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_image_prompt(image_url, prompt):
            async with semaphore:
                return await self.aprocess_image(
                    image_url=image_url, prompt=prompt, **kwargs
                )

        tasks = [
            process_image_prompt(image_url, prompt)
            for image_url, prompt in zip(image_urls, prompts)
        ]
        return await asyncio.gather(*tasks)

    @staticmethod
    def upload_file(file_path: str) -> str:
        """Upload a file and return the URL."""
        return fal_client.upload_file(file_path)



from typing import Any, Tuple, Dict, Optional, Union, Literal
from pydantic import Field, ConfigDict
from swarmauri_core.typing import SubclassUnion
from swarmauri.tools.base.ToolBase import ToolBase
from swarmauri_core.ComponentBase import ComponentBase, ResourceTypes
from swarmauri_core.chains.IChainStep import IChainStep

class ChainStepBase(IChainStep, ComponentBase):
    """
    Represents a single step within an execution chain.
    """
    key: str
    method: SubclassUnion[ToolBase]
    args: Tuple = Field(default_factory=tuple)
    kwargs: Dict[str, Any] = Field(default_factory=dict)
    ref: Optional[str] =  Field(default=None)
    resource: Optional[str] =  Field(default=ResourceTypes.CHAINSTEP.value)
    type: Literal['ChainStepBase'] = 'ChainStepBase'


import asyncio
import json
import logging
from time import sleep
from typing import List, Literal, Dict, Any
import mistralai
from swarmauri.conversations.concrete import Conversation
from swarmauri_core.typing import SubclassUnion

from swarmauri.messages.base.MessageBase import MessageBase
from swarmauri.messages.concrete.AgentMessage import AgentMessage
from swarmauri.messages.concrete.FunctionMessage import FunctionMessage
from swarmauri.llms.base.LLMBase import LLMBase
from swarmauri.schema_converters.concrete.MistralSchemaConverter import (
    MistralSchemaConverter,
)


class MistralToolModel(LLMBase):
    """
    Provider resources: https://docs.mistral.ai/capabilities/function_calling/#available-models
    """

    api_key: str
    allowed_models: List[str] = [
        "open-mixtral-8x22b",
        "mistral-small-latest",
        "mistral-large-latest",
        "open-mistral-nemo",
    ]
    name: str = "open-mixtral-8x22b"
    type: Literal["MistralToolModel"] = "MistralToolModel"

    def _schema_convert_tools(self, tools) -> List[Dict[str, Any]]:
        return [MistralSchemaConverter().convert(tools[tool]) for tool in tools]

    def _format_messages(
        self, messages: List[SubclassUnion[MessageBase]]
    ) -> List[Dict[str, str]]:
        message_properties = ["content", "role", "name", "tool_call_id"]
        # message_properties = ['content', 'role', 'tool_call_id', 'tool_calls']
        formatted_messages = [
            message.model_dump(include=message_properties, exclude_none=True)
            for message in messages
            if message.role != "assistant"
        ]
        logging.info(formatted_messages)
        return formatted_messages

    def predict(
        self,
        conversation,
        toolkit=None,
        tool_choice=None,
        temperature=0.7,
        max_tokens=1024,
        safe_prompt: bool = False,
    ):
        client = mistralai.Mistral(api_key=self.api_key)
        formatted_messages = self._format_messages(conversation.history)

        if toolkit and not tool_choice:
            tool_choice = "auto"
        tool_response = client.chat.complete(
            model=self.name,
            messages=formatted_messages,
            temperature=temperature,
            max_tokens=max_tokens,
            tools=self._schema_convert_tools(toolkit.tools),
            tool_choice=tool_choice,
            safe_prompt=safe_prompt,
        )

        logging.info(f"tool_response: {tool_response}")

        messages = [formatted_messages[-1], tool_response.choices[0].message]
        tool_calls = tool_response.choices[0].message.tool_calls
        if tool_calls:
            for tool_call in tool_calls:
                logging.info(type(tool_call.function.arguments))
                logging.info(tool_call.function.arguments)

                func_name = tool_call.function.name
                func_call = toolkit.get_tool_by_name(func_name)
                func_args = json.loads(tool_call.function.arguments)
                func_result = func_call(**func_args)

                messages.append(
                    {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": func_name,
                        "content": json.dumps(func_result),
                    }
                )
        logging.info(f"messages: {messages}")

        agent_response = client.chat.complete(model=self.name, messages=messages)
        logging.info(f"agent_response: {agent_response}")
        agent_message = AgentMessage(content=agent_response.choices[0].message.content)
        conversation.add_message(agent_message)
        logging.info(f"conversation: {conversation}")
        return conversation

    async def apredict(
        self,
        conversation,
        toolkit=None,
        tool_choice=None,
        temperature=0.7,
        max_tokens=1024,
        safe_prompt: bool = False,
    ):
        client = mistralai.Mistral(api_key=self.api_key)
        formatted_messages = self._format_messages(conversation.history)

        if toolkit and not tool_choice:
            tool_choice = "auto"

        tool_response = await client.chat.complete_async(
            model=self.name,
            messages=formatted_messages,
            temperature=temperature,
            max_tokens=max_tokens,
            tools=self._schema_convert_tools(toolkit.tools),
            tool_choice=tool_choice,
            safe_prompt=safe_prompt,
        )

        logging.info(f"tool_response: {tool_response}")

        messages = [formatted_messages[-1], tool_response.choices[0].message]
        tool_calls = tool_response.choices[0].message.tool_calls
        if tool_calls:
            for tool_call in tool_calls:
                logging.info(type(tool_call.function.arguments))
                logging.info(tool_call.function.arguments)

                func_name = tool_call.function.name
                func_call = toolkit.get_tool_by_name(func_name)
                func_args = json.loads(tool_call.function.arguments)
                func_result = func_call(**func_args)

                messages.append(
                    {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": func_name,
                        "content": json.dumps(func_result),
                    }
                )
        logging.info(f"messages: {messages}")

        agent_response = await client.chat.complete_async(
            model=self.name, messages=messages
        )
        logging.info(f"agent_response: {agent_response}")
        agent_message = AgentMessage(content=agent_response.choices[0].message.content)
        conversation.add_message(agent_message)
        logging.info(f"conversation: {conversation}")
        return conversation

    def stream(
        self,
        conversation,
        toolkit=None,
        tool_choice=None,
        temperature=0.7,
        max_tokens=1024,
        safe_prompt: bool = False,
    ):
        client = mistralai.Mistral(api_key=self.api_key)
        formatted_messages = self._format_messages(conversation.history)

        if toolkit and not tool_choice:
            tool_choice = "auto"

        tool_response = client.chat.complete(
            model=self.name,
            messages=formatted_messages,
            temperature=temperature,
            max_tokens=max_tokens,
            tools=self._schema_convert_tools(toolkit.tools),
            tool_choice=tool_choice,
            safe_prompt=safe_prompt,
        )

        logging.info(f"tool_response: {tool_response}")

        messages = [formatted_messages[-1], tool_response.choices[0].message]
        tool_calls = tool_response.choices[0].message.tool_calls

        if tool_calls:
            for tool_call in tool_calls:
                logging.info(type(tool_call.function.arguments))
                logging.info(tool_call.function.arguments)

                func_name = tool_call.function.name
                func_call = toolkit.get_tool_by_name(func_name)
                func_args = json.loads(tool_call.function.arguments)
                func_result = func_call(**func_args)

                messages.append(
                    {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": func_name,
                        "content": json.dumps(func_result),
                    }
                )
        logging.info(f"messages: {messages}")

        stream_response = client.chat.stream(model=self.name, messages=messages)
        message_content = ""

        for chunk in stream_response:
            if chunk.data.choices[0].delta.content:
                message_content += chunk.data.choices[0].delta.content
                yield chunk.data.choices[0].delta.content

        conversation.add_message(AgentMessage(content=message_content))

    async def astream(
        self,
        conversation,
        toolkit=None,
        tool_choice=None,
        temperature=0.7,
        max_tokens=1024,
        safe_prompt: bool = False,
    ):
        client = mistralai.Mistral(api_key=self.api_key)
        formatted_messages = self._format_messages(conversation.history)

        if toolkit and not tool_choice:
            tool_choice = "auto"

        tool_response = await client.chat.complete_async(
            model=self.name,
            messages=formatted_messages,
            temperature=temperature,
            max_tokens=max_tokens,
            tools=self._schema_convert_tools(toolkit.tools),
            tool_choice=tool_choice,
            safe_prompt=safe_prompt,
        )

        logging.info(f"tool_response: {tool_response}")

        messages = [formatted_messages[-1], tool_response.choices[0].message]
        tool_calls = tool_response.choices[0].message.tool_calls
        if tool_calls:
            for tool_call in tool_calls:
                logging.info(type(tool_call.function.arguments))
                logging.info(tool_call.function.arguments)

                func_name = tool_call.function.name
                func_call = toolkit.get_tool_by_name(func_name)
                func_args = json.loads(tool_call.function.arguments)
                func_result = func_call(**func_args)

                messages.append(
                    {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": func_name,
                        "content": json.dumps(func_result),
                    }
                )
        logging.info(f"messages: {messages}")

        stream_response = await client.chat.stream_async(
            model=self.name, messages=messages
        )
        message_content = ""

        async for chunk in stream_response:
            await asyncio.sleep(0.2) # 🚧 this is not an ideal permanent fix
            if chunk.data.choices[0].delta.content:
                message_content += chunk.data.choices[0].delta.content
                yield chunk.data.choices[0].delta.content

        conversation.add_message(AgentMessage(content=message_content))

    def batch(
        self,
        conversations: List[Conversation],
        toolkit=None,
        tool_choice=None,
        temperature=0.7,
        max_tokens=1024,
        safe_prompt: bool = False,
    ) -> List:
        """Synchronously process multiple conversations"""
        return [
            self.predict(
                conv,
                toolkit=toolkit,
                tool_choice=tool_choice,
                temperature=temperature,
                max_tokens=max_tokens,
                safe_prompt=safe_prompt,
            )
            for conv in conversations
        ]

    async def abatch(
        self,
        conversations: List[Conversation],
        toolkit=None,
        tool_choice=None,
        temperature=0.7,
        max_tokens=1024,
        safe_prompt: bool = False,
        max_concurrent: int = 5,
    ) -> List:
        """Process multiple conversations in parallel with controlled concurrency"""
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_conversation(conv):
            async with semaphore:
                return await self.apredict(
                    conv,
                    toolkit=toolkit,
                    tool_choice=tool_choice,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    safe_prompt=safe_prompt,
                )

        tasks = [process_conversation(conv) for conv in conversations]
        return await asyncio.gather(*tasks)



import asyncio
import json
from swarmauri.conversations.concrete.Conversation import Conversation
from typing import List, Optional, Dict, Literal, Any, Union, AsyncGenerator

from groq import Groq, AsyncGroq
from swarmauri_core.typing import SubclassUnion

from swarmauri.messages.base.MessageBase import MessageBase
from swarmauri.messages.concrete.AgentMessage import AgentMessage
from swarmauri.llms.base.LLMBase import LLMBase

from swarmauri.messages.concrete.AgentMessage import UsageData


class GroqModel(LLMBase):
    """Provider resources: https://console.groq.com/docs/models"""

    api_key: str
    allowed_models: List[str] = [
        "gemma-7b-it",
        "gemma2-9b-it",
        "llama-3.1-70b-versatile",
        "llama-3.1-8b-instant",
        "llama-3.2-11b-text-preview",
        "llama-3.2-1b-preview",
        "llama-3.2-3b-preview",
        "llama-3.2-90b-text-preview",
        "llama-guard-3-8b",
        "llama3-70b-8192",
        "llama3-8b-8192",
        "llama3-groq-70b-8192-tool-use-preview",
        "llama3-groq-8b-8192-tool-use-preview",
        "llava-v1.5-7b-4096-preview",
        "mixtral-8x7b-32768",
        # multimodal models
        "llama-3.2-11b-vision-preview",
    ]
    name: str = "gemma-7b-it"
    type: Literal["GroqModel"] = "GroqModel"

    def _format_messages(
        self,
        messages: List[SubclassUnion[MessageBase]],
    ) -> List[Dict[str, Any]]:
        formatted_messages = []
        for message in messages:
            formatted_message = message.model_dump(
                include=["content", "role", "name"], exclude_none=True
            )

            if isinstance(formatted_message["content"], list):
                formatted_message["content"] = [
                    {"type": item["type"], **item}
                    for item in formatted_message["content"]
                ]

            formatted_messages.append(formatted_message)
        return formatted_messages

    def _prepare_usage_data(
        self,
        usage_data,
    ):
        """
        Prepares and extracts usage data and response timing.
        """

        usage = UsageData.model_validate(usage_data)
        return usage

    def predict(
        self,
        conversation,
        temperature: float = 0.7,
        max_tokens: int = 256,
        top_p: float = 1.0,
        enable_json: bool = False,
        stop: Optional[List[str]] = None,
    ):

        formatted_messages = self._format_messages(conversation.history)
        response_format = {"type": "json_object"} if enable_json else None

        kwargs = {
            "model": self.name,
            "messages": formatted_messages,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "top_p": top_p,
            "response_format": response_format,
            "stop": stop or [],
        }

        client = Groq(api_key=self.api_key)
        response = client.chat.completions.create(**kwargs)

        result = json.loads(response.model_dump_json())
        message_content = result["choices"][0]["message"]["content"]
        usage_data = result.get("usage", {})

        usage = self._prepare_usage_data(usage_data)

        conversation.add_message(AgentMessage(content=message_content, usage=usage))
        return conversation

    async def apredict(
        self,
        conversation,
        temperature: float = 0.7,
        max_tokens: int = 256,
        top_p: float = 1.0,
        enable_json: bool = False,
        stop: Optional[List[str]] = None,
    ) -> Union[str, AsyncGenerator[str, None]]:

        formatted_messages = self._format_messages(conversation.history)
        response_format = {"type": "json_object"} if enable_json else None

        kwargs = {
            "model": self.name,
            "messages": formatted_messages,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "top_p": top_p,
            "response_format": response_format,
            "stop": stop or [],
        }

        client = AsyncGroq(api_key=self.api_key)
        response = await client.chat.completions.create(**kwargs)

        result = json.loads(response.model_dump_json())
        message_content = result["choices"][0]["message"]["content"]
        usage_data = result.get("usage", {})

        usage = self._prepare_usage_data(usage_data)

        conversation.add_message(AgentMessage(content=message_content, usage=usage))
        return conversation

    def stream(
        self,
        conversation,
        temperature: float = 0.7,
        max_tokens: int = 256,
        top_p: float = 1.0,
        enable_json: bool = False,
        stop: Optional[List[str]] = None,
    ) -> Union[str, AsyncGenerator[str, None]]:

        formatted_messages = self._format_messages(conversation.history)
        response_format = {"type": "json_object"} if enable_json else None

        kwargs = {
            "model": self.name,
            "messages": formatted_messages,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "top_p": top_p,
            "response_format": response_format,
            "stream": True,
            "stop": stop or [],
            # "stream_options": {"include_usage": True},
        }

        client = Groq(api_key=self.api_key)
        stream = client.chat.completions.create(**kwargs)
        message_content = ""
        # usage_data = {}

        for chunk in stream:
            if chunk.choices and chunk.choices[0].delta.content:
                message_content += chunk.choices[0].delta.content
                yield chunk.choices[0].delta.content

                # if hasattr(chunk, "usage") and chunk.usage is not None:
                #     usage_data = chunk.usage

        # usage = self._prepare_usage_data(usage_data)

        conversation.add_message(AgentMessage(content=message_content))

    async def astream(
        self,
        conversation,
        temperature: float = 0.7,
        max_tokens: int = 256,
        top_p: float = 1.0,
        enable_json: bool = False,
        stop: Optional[List[str]] = None,
    ) -> AsyncGenerator[str, None]:

        formatted_messages = self._format_messages(conversation.history)
        response_format = {"type": "json_object"} if enable_json else None

        kwargs = {
            "model": self.name,
            "messages": formatted_messages,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "top_p": top_p,
            "response_format": response_format,
            "stop": stop or [],
            "stream": True,
            # "stream_options": {"include_usage": True},
        }

        client = AsyncGroq(api_key=self.api_key)
        stream = await client.chat.completions.create(**kwargs)

        message_content = ""
        # usage_data = {}

        async for chunk in stream:
            if chunk.choices and chunk.choices[0].delta.content:
                message_content += chunk.choices[0].delta.content
                yield chunk.choices[0].delta.content

                # if hasattr(chunk, "usage") and chunk.usage is not None:
                #     usage_data = chunk.usage

        # usage = self._prepare_usage_data(usage_data)
        conversation.add_message(AgentMessage(content=message_content))

    def batch(
        self,
        conversations: List[Conversation],
        temperature: float = 0.7,
        max_tokens: int = 256,
        top_p: float = 1.0,
        enable_json: bool = False,
        stop: Optional[List[str]] = None,
    ) -> List:
        """Synchronously process multiple conversations"""
        return [
            self.predict(
                conv,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                enable_json=enable_json,
                stop=stop,
            )
            for conv in conversations
        ]

    async def abatch(
        self,
        conversations: List[Conversation],
        temperature: float = 0.7,
        max_tokens: int = 256,
        top_p: float = 1.0,
        enable_json: bool = False,
        stop: Optional[List[str]] = None,
        max_concurrent=5,
    ) -> List:
        """Process multiple conversations in parallel with controlled concurrency"""
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_conversation(conv):
            async with semaphore:
                return await self.apredict(
                    conv,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    top_p=top_p,
                    enable_json=enable_json,
                    stop=stop,
                )

        tasks = [process_conversation(conv) for conv in conversations]
        return await asyncio.gather(*tasks)



from typing import Any, Optional, Literal
from pydantic import BaseModel, ConfigDict, Field
from swarmauri_core.ComponentBase import ComponentBase, ResourceTypes
from swarmauri_core.measurements.IMeasurement import IMeasurement

class MeasurementBase(IMeasurement, ComponentBase):
    """
    A base implementation of the IMeasurement interface that provides the foundation
    for specific measurement implementations.
    """
    unit: str
    value: Any = None
    resource: Optional[str] =  Field(default=ResourceTypes.MEASUREMENT.value, frozen=True)
    type: Literal['MeasurementBase'] = 'MeasurementBase'

    def __call__(self, **kwargs) -> Any:
        """
        Retrieves the current value of the measurement.

        Returns:
            The current value of the measurement.
        """
        return self.value



import requests
import base64
from typing import List, Literal
from pydantic import Field
from swarmauri.llms.base.LLMBase import LLMBase
import asyncio
from typing import ClassVar


class DeepInfraImgGenModel(LLMBase):
    """
    A model for generating images from text using DeepInfra's image generation models.
    Resource: https://deepinfra.com/models/text-to-image/
    """

    api_key: str
    allowed_models: List[str] = [
        "black-forest-labs/FLUX-1-dev",
        "black-forest-labs/FLUX-1-schnell",
        "stabilityai/sdxl-turbo",
        "stabilityai/stable-diffusion-2-1",
    ]

    asyncio: ClassVar = asyncio
    name: str = "stabilityai/stable-diffusion-2-1"  # Default model
    type: Literal["DeepInfraImgGenModel"] = "DeepInfraImgGenModel"

    def _send_request(self, prompt: str) -> dict:
        """Send a request to DeepInfra's API for image generation."""
        url = f"https://api.deepinfra.com/v1/inference/{self.name}"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }
        data = {"prompt": prompt}

        response = requests.post(url, headers=headers, json=data)
        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(f"Error: {response.status_code}, {response.text}")

    def generate_image_base64(self, prompt: str) -> str:
        """Generates an image based on the prompt and returns the base64-encoded string."""
        # Send request to DeepInfra API
        response_data = self._send_request(prompt)

        # Extract the base64 image (the part after the data type prefix)
        image_base64 = response_data["images"][0].split(",")[1]

        return image_base64

    async def agenerate_image_base64(self, prompt: str) -> str:
        """Asynchronously generates an image based on the prompt and returns the base64-encoded string."""
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self.generate_image_base64, prompt)

    def batch_base64(self, prompts: List[str]) -> List[str]:
        """
        Generates base64-encoded images for a batch of prompts.
        Returns a list of base64 strings.
        """
        base64_images = []
        for prompt in prompts:
            base64_images.append(self.generate_image_base64(prompt=prompt))

        return base64_images

    async def abatch_base64(
        self, prompts: List[str], max_concurrent: int = 5
    ) -> List[str]:
        """
        Asynchronously generates base64-encoded images for a batch of prompts.
        Returns a list of base64 strings.
        """
        semaphore = asyncio.Semaphore(max_concurrent)

        async def process_prompt(prompt):
            async with semaphore:
                return await self.agenerate_image_base64(prompt=prompt)

        tasks = [process_prompt(prompt) for prompt in prompts]
        return await asyncio.gather(*tasks)



from swarmauri.agents.concrete.SimpleConversationAgent import SimpleConversationAgent
from swarmauri.agents.concrete.QAAgent import QAAgent
from swarmauri.agents.concrete.RagAgent import RagAgent
from swarmauri.agents.concrete.ToolAgent import ToolAgent

__all__ = [
    "SimpleConversationAgent",
    "QAAgent",
    "RagAgent",
    "ToolAgent",
]



from typing import Any, Optional, Union, Dict, Literal
from swarmauri_core.messages import IMessage

from swarmauri.agents.base.AgentBase import AgentBase
from swarmauri.agents.base.AgentRetrieveMixin import AgentRetrieveMixin
from swarmauri.agents.base.AgentConversationMixin import AgentConversationMixin
from swarmauri.agents.base.AgentVectorStoreMixin import AgentVectorStoreMixin
from swarmauri.agents.base.AgentSystemContextMixin import AgentSystemContextMixin

from swarmauri.messages.concrete import (HumanMessage, 
                                                  SystemMessage,
                                                  AgentMessage)

from swarmauri_core.typing import SubclassUnion
from swarmauri.llms.base.LLMBase import LLMBase
from swarmauri.conversations.base.ConversationBase import ConversationBase
from swarmauri.vector_stores.base.VectorStoreBase import VectorStoreBase

class RagAgent(AgentRetrieveMixin, 
               AgentVectorStoreMixin, 
               AgentSystemContextMixin, 
               AgentConversationMixin, 
               AgentBase):
    """
    RagAgent (Retriever-And-Generator Agent) extends DocumentAgentBase,
    specialized in retrieving documents based on input queries and generating responses.
    """
    llm: SubclassUnion[LLMBase]
    conversation: SubclassUnion[ConversationBase]
    vector_store: SubclassUnion[VectorStoreBase]
    system_context:  Union[SystemMessage, str]
    type: Literal['RagAgent'] = 'RagAgent'
    
    def _create_preamble_context(self):
        substr = self.system_context.content
        substr += '\n\n'
        substr += '\n'.join([doc.content for doc in self.last_retrieved])
        return substr

    def _create_post_context(self):
        substr = '\n'.join([doc.content for doc in self.last_retrieved])
        substr += '\n\n'
        substr += self.system_context.content
        return substr

    def exec(self, 
             input_data: Optional[Union[str, IMessage]] = "", 
             top_k: int = 5, 
             preamble: bool = True,
             fixed: bool = False,
             llm_kwargs: Optional[Dict] = {}
             ) -> Any:
        try:
            # Check if the input is a string, then wrap it in a HumanMessage
            if isinstance(input_data, str):
                human_message = HumanMessage(content=input_data)
            elif isinstance(input_data, IMessage):
                human_message = input_data
            else:
                raise TypeError("Input data must be a string or an instance of Message.")
            
            # Add the human message to the conversation
            self.conversation.add_message(human_message)

            # Retrieval and set new substr for system context
            if top_k > 0 and len(self.vector_store.documents) > 0:
                self.last_retrieved = self.vector_store.retrieve(query=input_data, top_k=top_k)

                if preamble:
                    substr = self._create_preamble_context()
                else:
                    substr = self._create_post_context()

            else:
                if fixed:
                    if preamble:
                        substr = self._create_preamble_context()
                    else:
                        substr = self._create_post_context()
                else:
                    substr = self.system_context.content
                    self.last_retrieved = []
                
            # Use substr to set system context
            system_context = SystemMessage(content=substr)
            self.conversation.system_context = system_context
            

            # Retrieve the conversation history and predict a response
            if llm_kwargs:
                self.llm.predict(conversation=self.conversation, **llm_kwargs)
            else:
                self.llm.predict(conversation=self.conversation)
                
            return self.conversation.get_last().content

        except Exception as e:
            print(f"RagAgent error: {e}")
            raise e


from abc import ABC, abstractmethod
from typing import Any, Union, Optional, List, Literal
from pydantic import BaseModel, ConfigDict, ValidationError, model_validator, Field
from swarmauri_core.ComponentBase import ComponentBase, ResourceTypes
from swarmauri_core.llms.IPredict import IPredict


class LLMBase(IPredict, ComponentBase):
    allowed_models: List[str] = []
    resource: Optional[str] = Field(default=ResourceTypes.LLM.value, frozen=True)
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)
    type: Literal["LLMBase"] = "LLMBase"

    @model_validator(mode="after")
    @classmethod
    def _validate_name_in_allowed_models(cls, values):
        name = values.name
        allowed_models = values.allowed_models
        if name and name not in allowed_models:
            raise ValueError(
                f"Model name {name} is not allowed. Choose from {allowed_models}"
            )
        return values

    def predict(self, *args, **kwargs):
        raise NotImplementedError("predict() not implemented in subclass yet.")

    async def apredict(self, *args, **kwargs):
        raise NotImplementedError("apredict() not implemented in subclass yet.")

    def stream(self, *args, **kwargs):
        raise NotImplementedError("stream() not implemented in subclass yet.")

    async def astream(self, *args, **kwargs):
        raise NotImplementedError("astream() not implemented in subclass yet.")

    def batch(self, *args, **kwargs):
        raise NotImplementedError("batch() not implemented in subclass yet.")

    async def abatch(self, *args, **kwargs):
        raise NotImplementedError("abatch() not implemented in subclass yet.")